using System;
using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Threading;


namespace TestSpeed {
    class FindDens {
        const int MAX_NP=16000000;
        int m_dim;
        int m_np;
        ulong[] m_Arr;
        static ulong MAXBIT=(ulong.MaxValue>>1)+1;
        ulong[] m_PackCvt;
        int m_Coeff,m_Shift;

        public FindDens(int _dim) {
            m_Arr=new ulong[MAX_NP];
            m_dim=_dim;
            m_np=0;

            int nb=31/_dim+1;
            m_Coeff=1<<nb;
            m_PackCvt=new ulong[m_Coeff];
// (nb-1) -> 63
            for(int k=0;k<nb;k++) {
                int m=1<<k;
                ulong l=MAXBIT>>((nb-1-k)*m_dim);
                for(int i=0;i<m_Coeff;i++) m_PackCvt[i|m]|=l;
            }
            m_Shift=nb*m_dim;
        }

        ulong Pack(double[] _arr) {
            ulong res=0;
            for(int i=0;i<m_dim;i++) {
                double a=_arr[i]*m_Coeff;
                int v=(int)a; a-=v;
                res|=m_PackCvt[v]>>i;
                res|=m_PackCvt[(int)(a*m_Coeff)]>>(i+m_Shift);
            }
            return res;
        }
        double[] Expand(ulong v) {
            double[] res=new double[m_dim];
            double am=0.5;
            ulong bit=MAXBIT;
            int h=0;
            while(bit!=0) {
                if((v&bit)!=0) res[h]+=am;
                bit>>=1; h++;
                if(h==m_dim) { h=0; am/=2; }
            }
            return res;
        }
        public void AddPoint(double[] _pt) {
            if(m_np==MAX_NP) return;
            m_Arr[m_np++]=Pack(_pt);
        }
        public double[] GetBestArea(out double _xdens) {
            _xdens=0;
            if(m_np==0) return null;
            Array.Sort<ulong>(m_Arr,0,m_np);
            int ms,me,rank;
            GetPrimArea((int)Math.Sqrt(m_np),out ms,out me,out rank);
            _xdens=(me-ms+1)*Math.Pow(2,rank);
            while(m_Arr[ms]!=m_Arr[me]) ImproveArea(ref ms,ref me);
            return Expand(m_Arr[ms]);
        }
        private void GetPrimArea(int np,out int ms,out int me,out int rank) {
            ulong mask=ulong.MaxValue;
            np--;
            for(int i=np;i<m_np;i++) mask=Math.Min(mask,m_Arr[i]^m_Arr[i-np]);
            rank=0;
            for(ulong b=MAXBIT;b!=0 && (mask&b)==0;b>>=1) rank++;
            mask=InvMask(mask);
            ms=me=0;
            for(int i=0;i<m_np-np;i++){
                ulong a=m_Arr[i];
                int h=i+np;
                if(((a^m_Arr[h])&mask)==0){
                    while(h<m_np-1 && ((a^m_Arr[h+1])&mask)==0) h++;
                    np=h-i; 
                    ms=i; me=i=h;
                }
            }
        }
        private void ImproveArea(ref int ms,ref int me) {
            int h=(ms+me)/2;
            ulong samp=m_Arr[h];
            ulong cb=samp^m_Arr[ms],ce=samp^m_Arr[me];
            if(cb>ce) {
                ms=h;
                ce=InvMask(ce);
                while(ms>0 && ((m_Arr[ms-1]^samp)&ce)==0) ms--;
            } else {
                me=h;
                cb=InvMask(cb);
                while(me<m_np-1 && ((m_Arr[me+1]^samp)&cb)==0) me++;
            }
        }
        private static ulong InvMask(ulong cb) {
            while((cb&(cb+1))!=0) cb|=cb+1; // keep it simple and stupid
            return ~cb;
        }
    }

    class TestDens {
        static internal void Test() {
            Random rRand = new Random();
            FindDens fdens=new FindDens(3);
            double[] pt=new double[3];
            long t1=DateTime.Now.Ticks;
            for(int i=0;i<10000000;i++) {
                pt[0]=func(rRand.NextDouble());
                pt[1]=func(rRand.NextDouble());
                pt[2]=func(rRand.NextDouble());
                fdens.AddPoint(pt);
            }
            long t3=DateTime.Now.Ticks;
            double dens;
            double[] res=fdens.GetBestArea(out dens);
            long t4=DateTime.Now.Ticks;
            Console.WriteLine("Preparation+Fill: {0:F3}s",(t3-t1)*1e-7);
            Console.WriteLine("Compute: {0:F3}s",(t4-t3)*1e-7);            
            Console.WriteLine("({0},{1},{2}), dens={3}",res[0],res[1],res[2],dens);
        }
        static double func(double x) {
            return (3*(x-1)*x+1)*x;
        }
    }
    class Prog{
        public static void Main(string[] args) {
            TestDens.Test();
        }
    }
}
